from pystac.provider import Provider, ProviderRole

COLLECTION_ID = 'noaa-c-cap'
COLLECTION_TITLE = 'C-CAP Regional Land Cover and Change'
COLLECTION_KEYWORDS = ['landcover', 'coastal', 'change']
COLLECTION_PROVIDERS = [
    Provider(name=('National Oceanic and Atmospheric Administration, '
                   'Office for Coastal Management'),
             roles=[
                 ProviderRole.LICENSOR, ProviderRole.PRODUCER,
                 ProviderRole.PROCESSOR, ProviderRole.HOST
             ],
             url='https://coast.noaa.gov/digitalcoast/data/ccapregional.html')
]
COLLECTION_DESCRIPTION = """Nationally standardized, raster-based inventories of
land cover for the coastal areas of the U.S.  Data are derived, through the
Coastal Change Analysis Program, from the analysis of multiple dates of remotely
sensed imagery.  Two file types are available: individual dates that supply a
wall-to-wall map, and change files that compare one date to another.

The use of standardized data and procedures assures consistency through time and
across geographies.  C-CAP data forms the coastal expression of the National
Land Cover Database (NLCD) and the A-16 land cover theme of the National Spatial
Data Infrastructure.  The data are updated every 5 years."""
LABEL_CLASSES = [
    "Background", "Unclassified (Cloud, Shadow, etc)",
    "High Intensity Developed", "Medium Intensity Developed",
    "Low Intensity Developed", "Developed Open Space", "Cultivated Land",
    "Pasture/Hay", "Grassland", "Deciduous Forest", "Evergreen Forest",
    "Mixed Forest", "Scrub/Shrub", "Palustrine Forested Wetland",
    "Palustrine Scrub/Shrub Wetland", "Palustrine Emergent Wetland",
    "Estuarine Forested Wetland", "Estuarine Scrub/Shrub Wetland",
    "Estuarine Emergent Wetland", "Unconsolidated Shore", "Bare Land",
    "Open Water", "Palustrine Aquatic Bed", "Estuarine Aquatic Bed", "Tundra",
    "Snow/Ice"
]
COLLECTION_CITATION = (
    "National Oceanic and Atmospheric Administration, "
    "Office for Coastal Management. "
    "Name of Data Set: Coastal Change Analysis Program (C-CAP) Regional Land Cover. "
    "Charleston, SC: NOAA Office for Coastal Management. "
    "Accessed Month Year at www.coast.noaa.gov/htdata/raster1/landcover/bulkdownload/30m_lc/."
)



import stactools.core

from stactools.noaa_c_cap.dataset import Dataset

stactools.core.use_fsspec()


def register_plugin(registry):
    from stactools.noaa_c_cap import commands
    registry.register_subcommand(commands.create_noaa_c_cap_command)


__all__ = ['Dataset']
__version__ = "0.1.3"



import logging
import os
from tempfile import TemporaryDirectory
from typing import Optional, Tuple

import click
import requests
import stactools.core.utils.convert
from click.types import Choice
from pystac.catalog import CatalogType

from stactools.noaa_c_cap import stac, utils

logger = logging.getLogger(__name__)


def create_noaa_c_cap_command(cli):
    """Creates the stactools-noaa-c-cap command line utility."""

    @cli.group(
        "noaa-c-cap",
        short_help=("Commands for working with stactools-noaa-c-cap"),
    )
    def noaa_c_cap():
        pass

    @noaa_c_cap.command(
        "create-collection",
        short_help="Creates a STAC collection",
    )
    @click.argument("destination")
    @click.option('--href', '-h', multiple=True)
    @click.option('--directory',
                  '-d',
                  help='Directory containing NOAA C-CAP files')
    @click.option('--catalog-type',
                  type=Choice([
                      'ABSOLUTE_PUBLISHED', 'RELATIVE_PUBLISHED',
                      'SELF_CONTAINED'
                  ]),
                  default='ABSOLUTE_PUBLISHED')
    def create_collection_command(destination: str, href: Tuple[str],
                                  directory: Optional[str],
                                  catalog_type: CatalogType):
        """Creates a STAC Collection
        Args:
            destination (str): An HREF for the Collection JSON
        """
        hrefs = list(href)
        if directory:
            hrefs.extend(
                os.path.join(os.path.abspath(directory), file_name)
                for file_name in os.listdir(directory))
        collection = stac.create_collection(hrefs=hrefs)
        collection.catalog_type = catalog_type
        collection.normalize_and_save(destination)

    @noaa_c_cap.command("create-item", short_help="Create a STAC item")
    @click.argument("source")
    @click.argument("destination")
    @click.option('-x', '--xml', help='HREF to the metadata XML file.')
    @click.option('--cogify/--no-cogify',
                  default=False,
                  help=('Use a Cloud-Optimized GeoTIFF version '
                        'of the input file as the data asset.'))
    def create_item_command(source: str, destination: str, xml: Optional[str],
                            cogify: bool):
        """Creates a STAC Item

        Args:
            source (str): HREF of the Asset associated with the Item
            destination (str): An HREF for the STAC Collection
            xml (str): HREF to the metadata XML file
            cogify (bool): Use a Cloud-Optimized GeoTIFF version of the source
            file as the data asset
        """
        if cogify:
            cog = os.path.join(os.path.dirname(destination),
                               os.path.basename(source))
            stactools.core.utils.convert.cogify(source, cog)
            source = cog
        item = stac.create_item(source, xml)
        item.save_object(dest_href=destination)

    @noaa_c_cap.command(
        "download",
        short_help="Download NOAA C-CAP data into a target directory")
    @click.argument("destination")
    @click.option('--cogify/--no-cogify',
                  help="Convert the .tifs to COGs after download",
                  default=False)
    def download_command(destination: str, cogify: bool) -> None:
        """Downloads the source data to a target directory.

        Args:
            destination (str): A local directory into which the data will downloaded.
        """
        os.makedirs(destination, exist_ok=True)
        with TemporaryDirectory() as temporary_directory:
            for url in utils.urls():
                file_name = os.path.basename(url)
                final_path = os.path.join(destination, file_name)
                print(f"Downloading {url} to {final_path}")
                download_to_temporary_directory = cogify and os.path.splitext(
                    file_name)[1] == '.tif'
                if download_to_temporary_directory:
                    download_path = os.path.join(temporary_directory,
                                                 file_name)
                else:
                    download_path = final_path
                with open(download_path, 'wb') as f:
                    response = requests.get(url, stream=True)
                    for chunk in response.iter_content(chunk_size=1024):
                        f.write(chunk)
                if download_to_temporary_directory:
                    stactools.core.utils.convert.cogify(
                        download_path, final_path)

    return noaa_c_cap



import logging
import os.path
from typing import List, Optional

logger = logging.getLogger(__name__)


class Dataset:
    """A NOAA C-CAP dataset.

    Could be CONUS, could be somewhere else. Always has a tiff file, can have XML metadata.
    """
    tiff_href: str
    year: str
    location: str
    xml_file_name: Optional[str]
    xml_href: Optional[str]
    classes: List[str]

    @classmethod
    def from_hrefs(cls, hrefs: List[str]) -> List['Dataset']:
        """Creates one or more datasets from a list of hrefs.

        Does the work of looking through the HREFs to associate tifs with xml metadata.
        """
        datasets = []
        for tif_href in (href for href in hrefs
                         if os.path.splitext(href)[1] == '.tif'):
            dataset = Dataset(tif_href)
            if dataset.xml_file_name:
                xml_href = next(
                    (href for href in hrefs
                     if os.path.basename(href) == dataset.xml_file_name), None)
                if xml_href:
                    dataset.xml_href = xml_href
                else:
                    logger.warn(
                        f"Could not find XML metadata file for {tif_href}")
            datasets.append(dataset)
        return datasets

    def __init__(self, tiff_href: str, xml_href: Optional[str] = None):
        """Creates a new dataset for a GeoTIFF."""
        self.tiff_href = tiff_href
        self.xml_href = xml_href

        file_name = os.path.basename(tiff_href)
        parts = file_name.split('_')
        if len(parts) < 2:
            raise ValueError(f"Invalid NOAA C-CAP file name: {file_name}")
        self.location = parts[0]
        self.year = parts[1]
        if self.location == 'conus':
            self.xml_file_name: Optional[str] = f"CCAP_Parent_{self.year}.xml"
        elif self.location == 'hi':
            self.xml_file_name = None
        elif self.location == 'pr':
            self.xml_file_name = f"{self.year}_puerto_rico_ccap.xml"



import datetime
import logging
from typing import List, Optional

from pystac import Asset, Collection, Extent, Item, MediaType, Summaries
from pystac.extensions.item_assets import AssetDefinition, ItemAssetsExtension
from pystac.extensions.label import LabelClasses, LabelExtension, LabelType
from pystac.extensions.scientific import ScientificExtension
from stactools.core import create
from stactools.core.io import ReadHrefModifier

from stactools.noaa_c_cap import Dataset, utils
from stactools.noaa_c_cap.constants import (COLLECTION_CITATION,
                                            COLLECTION_DESCRIPTION,
                                            COLLECTION_ID, COLLECTION_KEYWORDS,
                                            COLLECTION_PROVIDERS,
                                            COLLECTION_TITLE, LABEL_CLASSES)

logger = logging.getLogger(__name__)


def create_collection(hrefs: Optional[List[str]] = None) -> Collection:
    """Creates a STAC Collection for all C-CAP data.

    Args:
        hrefs (Optional[str]): If provided, create a Collection for the provided
        HREFs. If not provided, use the network locations defined in
        `urllist-modres.txt`.
    """
    if not hrefs:
        hrefs = utils.urls()
    datasets = Dataset.from_hrefs(hrefs)
    items = [create_item_from_dataset(dataset) for dataset in datasets]
    extent = Extent.from_items(items)
    summaries = Summaries.empty()
    summaries.add("gsd", list(set(item.common_metadata.gsd for item in items)))
    summaries.add("label:classes", [{"classes": LABEL_CLASSES}])
    collection = Collection(id=COLLECTION_ID,
                            title=COLLECTION_TITLE,
                            description=COLLECTION_DESCRIPTION,
                            extent=extent,
                            keywords=COLLECTION_KEYWORDS,
                            providers=COLLECTION_PROVIDERS,
                            summaries=summaries)
    collection.add_items(items)

    item_assets = {}
    for item in items:
        for key, asset in item.get_assets().items():
            asset_as_dict = asset.to_dict()
            asset_as_dict.pop('href')
            if key not in item_assets:
                item_assets[key] = AssetDefinition(asset_as_dict)
            elif item_assets[key].to_dict() != asset_as_dict:
                logger.warning(f"Item Asset '{key}' does not match asset:\n"
                               f"item_asset={item_assets[key].to_dict()}\n"
                               f"asset={asset_as_dict}")

    item_assets_ext = ItemAssetsExtension.ext(collection, add_if_missing=True)
    item_assets_ext.item_assets = item_assets

    scientific = ScientificExtension.ext(collection, add_if_missing=True)
    scientific.citation = COLLECTION_CITATION

    return collection


def create_item(tiff_href: str,
                xml_href: Optional[str] = None,
                read_href_modifier: Optional[ReadHrefModifier] = None) -> Item:
    """Creates a STAC Item for the provided C-CAP tiff file.

    Args:
        tiff_href (str): The HREF pointing to the tiff file.
        xml_href (str): The HREF pointing to the xml file.
        read_href_modifier (Callable[[str], str]): A function to modify the read href for the file.

    Returns:
        Item: STAC Item object
    """
    return create_item_from_dataset(Dataset(tiff_href=tiff_href,
                                            xml_href=xml_href),
                                    read_href_modifier=read_href_modifier)


def create_item_from_dataset(
        dataset: Dataset,
        read_href_modifier: Optional[ReadHrefModifier] = None) -> Item:
    """Creates an item from a NOAA C-CAP dataset."""
    logger.info(f"Creating STAC item from {dataset.tiff_href}")
    item = create.item(dataset.tiff_href,
                       read_href_modifier=read_href_modifier)
    item.common_metadata.start_datetime = datetime.datetime(
        int(dataset.year), 1, 1, tzinfo=datetime.timezone.utc)
    item.common_metadata.end_datetime = datetime.datetime(
        int(dataset.year), 12, 31, tzinfo=datetime.timezone.utc)
    item.common_metadata.gsd = 30
    item.datetime = item.common_metadata.start_datetime

    label = LabelExtension.ext(item, add_if_missing=True)
    label.label_properties = None
    label.label_description = "Land cover classes"
    label.label_type = LabelType.RASTER
    label.label_classes = [
        LabelClasses.create(name='classes', classes=list(LABEL_CLASSES))
    ]

    data = item.assets.get('data')
    assert data
    data.media_type = MediaType.GEOTIFF
    if dataset.xml_href:
        item.add_asset(
            'metadata',
            Asset(dataset.xml_href,
                  media_type=MediaType.XML,
                  roles=['metadata']))
    return item



from typing import List

import pkg_resources


def urls() -> List[str]:
    """Returns the list of URLs defined in urllist-modres.txt.

    These are URLs to _all_ of the NOAA C-CAP data and metadata files.
    """
    return pkg_resources.resource_string(
        __name__, 'urllist-modres.txt').decode('utf-8').splitlines()
